import { __assign } from "tslib";
import * as extractors from "../featureExtractors";
import { createBarkScale, createChromaFilterBank, createMelFilterBank, } from "../utilities";
import { fft } from "fftjs";
function configure(options) {
    var _a = __assign({}, options), sampleRate = _a.sampleRate, bufferSize = _a.bufferSize, melBands = _a.melBands, mfccCoefficients = _a.mfccCoefficients, chromaBands = _a.chromaBands;
    return __assign(__assign({}, options), { barkScale: createBarkScale(bufferSize, sampleRate, bufferSize), melFilterBank: createMelFilterBank(Math.max(melBands || 0, mfccCoefficients), sampleRate, bufferSize), chromaFilterBank: createChromaFilterBank(chromaBands, sampleRate, bufferSize) });
}
function prepareAmpSpectrum(complexSpectrum, bufferSize) {
    var ampSpectrum = new Float32Array(bufferSize / 2);
    for (var i = 0; i < bufferSize / 2; i++) {
        ampSpectrum[i] = Math.sqrt(Math.pow(complexSpectrum.real[i], 2) +
            Math.pow(complexSpectrum.imag[i], 2));
    }
    return ampSpectrum;
}
function prepareSpectra(signal) {
    var complexSpectrum = fft(signal);
    var ampSpectrum = prepareAmpSpectrum(complexSpectrum, signal.length);
    return {
        complexSpectrum: complexSpectrum,
        ampSpectrum: ampSpectrum
    };
}
function prepareExtractorDependencies(features, signal, configuration
// ): ExtractorParameters<T> {
) {
    var bufferSize = configuration.bufferSize, sampleRate = configuration.sampleRate, melFilterBank = configuration.melFilterBank, chromaFilterBank = configuration.chromaFilterBank, barkBands = configuration.barkBands, barkScale = configuration.barkScale, mfccCoefficients = configuration.mfccCoefficients;
    var _a = prepareSpectra(signal), complexSpectrum = _a.complexSpectrum, ampSpectrum = _a.ampSpectrum;
    var previousSignal = undefined;
    return {
        signal: !(signal instanceof Float32Array)
            ? Float32Array.from(signal)
            : signal,
        complexSpectrum: complexSpectrum,
        ampSpectrum: ampSpectrum,
        bufferSize: bufferSize,
        sampleRate: sampleRate,
        chromaFilterBank: chromaFilterBank,
        melFilterBank: melFilterBank,
        numberOfMFCCCoefficients: mfccCoefficients,
        numberOfBarkBands: barkBands,
        barkScale: barkScale,
        previousSignal: previousSignal
    };
}
/**
 * # Configure an extractor
 *
 * Configure an extractor function with options that are shared throughout the
 * extractor function's lifetime. The options pertain to the type of audio that
 * the extractor expects - e.g. the sample rate, the consistent windowing
 * function to apply, etc.
 *
 * @param options Configuration used across all feature extractions made with
 * resulting interface
 
 * @returns A function that takes a set of requested audio features and one
 * buffer of audio, and returns the requested audio features extracted from that
 * audio.
 */
export function configureMeyda(options) {
    var defaults = {
        bufferSize: 512,
        sampleRate: 44100,
        melBands: 26,
        chromaBands: 12,
        windowingFunction: "hanning",
        mfccCoefficients: 13,
        barkBands: 24
    };
    var optionsAfterDefaults = __assign(__assign({}, defaults), options);
    var configuration = configure(optionsAfterDefaults);
    /**
     * ## Extract given audio features from given signal
     *
     * @param feature The audio feature(s) to extract. Can be single or an array
     * @param signal The audio signal to extract the feature(s) from. Can be an
     * array containing each channel of a frame of multi-channel audio. If signal
     * is multi-channel, the result will be an array.
     *
     * The signal should have the same sample rate as specified in the options in
     * the call to `configureMeyda`. Otherwise, the audio features will be wrong.
     * There is no way to validate the sample rate of the signal, so you will get
     * mangled output if you don't match it.
     *
     * @returns The audio feature(s) extracted from the signal. If the signal is
     * multi-channel, this will be an array of results for each channel in the
     * same order that the channels were passed in.
     *
     * *NB: Multi-Channel -* Meyda saves a portion of the history of each channel
     * of audio for subsequent feature extraction runs. We assume that the
     * channels are passed in in the same order for every call to extract.
     * Changing the order of the channels between calls will produce invalid
     * results. We have no way to detect or warn you about this, so please make
     * sure that your audio channels are passed in in the same order every time.
     */
    function extract(feature, signal) {
        // const features: readonly T[] = isReadonlyArray<T>(feature as readonly T[] | T)
        //   ? feature
        //   : [feature];
        var features = (Array.isArray(feature)
            ? feature
            : [feature]);
        var signals = Array.isArray(signal[0])
            ? signal
            : [signal];
        var results = signals.map(function (signal) {
            // TODO:
            // apply windowing function
            // run fft on signal
            // cache previous (+?) signal, amp spectrum and complex spectrum per channel
            // run extractors with all deps
            // collect to MeydaExtractionResult
            var preparedExtractorDependencies = prepareExtractorDependencies(features, signal, configuration);
            var channelExtractionResult = Object.fromEntries(features.map(function (feature) {
                var extractor = extractors[feature];
                return [feature, extractors[feature](preparedExtractorDependencies)];
            }));
            return channelExtractionResult;
        });
        if (!Array.isArray(signal)) {
            return results[0];
        }
        return results;
    }
    return extract;
}
/**
 * Returns a Meyda extractor function with configured audio features.
 *
 * @param features
 * @param extractor
 * @returns
 */
export function curryMeyda(features, extractor) {
    return function extract(signal) {
        var result = extractor(features, signal);
        return result;
    };
}
/**
 * ### Configure an extractor with options and audio features
 *
 * Configure a Meyda extractor function with options and audio features at once.
 *
 * @param features
 * @param options
 * @returns
 */
export function configureMeydaWithExtractors(features, options) {
    return function extract(signal) {
        var extractor = configureMeyda(options);
        return curryMeyda(features, extractor)(signal);
    };
}
// type FType<
//   T extends V extends readonly (infer K)[] ? K : V extends (infer K)[] ? K : V,
//   U extends readonly MeydaSignal[] | MeydaSignal,
//   V extends
//     | MeydaAudioFeature[]
//     | readonly MeydaAudioFeature[]
//     | MeydaAudioFeature
// > = (feature: V, signal: U) => MaybePartialResult<T, U, V>;
